const NS_RFHELPER_CONTRACTID = "@roboform.org/rfhelper;3"
const NS_RFHELPER_CID = Components.ID("{68A9111B-10CF-4f2e-BFB2-E13A869F0003}");

function NOT_IMPLEMENTED() {
	throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
}

const CI = Components.interfaces;

const nsIComponentRegistrar     = CI.nsIComponentRegistrar;
const nsIFactory                = CI.nsIFactory;
const nsIModule                 = CI.nsIModule;
const nsISupports               = CI.nsISupports;
const nsIComponentManager		= CI.nsIComponentManager;
const nsIObserver				= CI.nsIObserver;
const nsIDOMWindowInternal		= CI.nsIDOMWindowInternal;
const nsIUnicharStreamListener    = CI.nsIUnicharStreamListener;

function RFUnicharStreamListener(handler) {
    this.onUnicharDataAvailable = handler;
    this.QueryInterface = function (iid) {
        if (iid.equals(nsISupports) ||
            iid.equals(nsIUnicharStreamListener))
            return this;
        throw Components.results.NS_ERROR_NO_INTERFACE;
    };
}
//    void onUnicharDataAvailable(in nsIRequest aRequest,
//                                in nsISupports aContext, in AString aData);

function RFComponentManager(handler) {
    this.createInstanceByContractID = handler;
    this.QueryInterface = function (iid) {
        if (iid.equals(nsISupports) ||
            iid.equals(nsIComponentManager))
            return this;
        throw Components.results.NS_ERROR_NO_INTERFACE;
    };
}
//    void createInstanceByContractID(in string aContractID,
//                                    in nsISupports aDelegate,
//                                    in nsIIDRef aIID, 
//                                    [iid_is(aIID),retval] out nsQIResult result);

function RFSimpleComponentManager(iid, prop) {
    return new RFComponentManager(
        function(sContractID, pDelegate, aIID) {
    	    return pDelegate.QueryInterface(iid)[opener];
    	}
    );
}

function RFLoadGroup(handler) {
    this.addRequest = handler;
    this.QueryInterface = function (iid) {
        if (iid.equals(nsISupports) ||
            iid.equals(CI.nsILoadGroup))
            return this;
        throw Components.results.NS_ERROR_NO_INTERFACE;
    };
}
//    void addRequest(in nsIRequest aRequest,
//                    in nsISupports aContext);

function RFSimpleLoadGroup(iid, iid_callback, prop) {
    return new RFLoadGroup(
        function(aRequest, aElem) {
            aRequest.QueryInterface(iid_callback).data = aElem.QueryInterface(iid)[prop];
    	}
    );
}

function RFProperties(handler) {
    this.set = handler;
    this.QueryInterface = function (iid) {
        if (iid.equals(nsISupports) ||
            iid.equals(CI.nsIProperties))
            return this;
        throw Components.results.NS_ERROR_NO_INTERFACE;
    };
}
//    void set(in string prop, in nsISupports value);

function RFCommonWraper(hname, handler, iid) {
    this.mIID = iid;
    this[hname] = handler;
    this.QueryInterface = function (iid) {
        if (iid.equals(nsISupports) ||
            iid.equals(this.mIID))
            return this;
        throw Components.results.NS_ERROR_NO_INTERFACE;
    };
}

function RFFactory(handler) {
    return new RFCommonWraper("createInstance", handler, CI.nsIFactory);
}
//    void createInstance(in nsISupports aOuter, in nsIIDRef iid,
//                      [retval, iid_is(iid)] out nsQIResult result);

function RFSimpleFactory(miid, prop) {
    return RFFactory(
        function(aOuter, iid) {
    	    return aOuter.QueryInterface(miid)[prop];
    	}
    );
}

function RFArray(handler) {
    return new RFCommonWraper("indexOf", handler, CI.nsIArray);
}
//    unsigned long indexOf(in unsigned long startIndex,
//                          in nsISupports element);

function RFSimpleArray(miid, prop) {
    return RFArray(
        function(ulIndex, elem) {
   	        return elem.QueryInterface(miid)[prop];
    	}
    );
}

function RFStreamListener(handler) {
    return new RFCommonWraper("onDataAvailable", handler, CI.nsIStreamListener);
}
//    void onDataAvailable(in nsIRequest aRequest,
//                         in nsISupports aContext, 
//                         in nsIInputStream aInputStream,
//                         in unsigned long aOffset,
//                         in unsigned long aCount);

const RFHelper = {
	// nsIFactory
	createInstance: function createInstance(outer, iid) {
		return this;
	},

    //nsIComponentManager    
    getClassObjectByContractID: function getClassObjectByContractID(sContractID, aIID) {
    	switch(sContractID) {
    	//ComponentManager
    	case "Window_GetOpener": return RFSimpleFactory(CI.nsIDOMWindowInternal, "opener");
    	case "Window_GetDocument": return RFSimpleFactory(CI.nsIDOMWindowInternal, "document");
    	case "Window_GetToolbar": return RFSimpleFactory(CI.nsIDOMWindowInternal, "toolbar");
    	case "Window_GetContent": return RFSimpleFactory(CI.nsIDOMWindowInternal, "content");
    	case "Window_GetFrameElement": return RFSimpleFactory(CI.nsIDOMWindowInternal, "frameElement");
    	case "Window_GetDocShell":
    	    return new RFFactory(
    	        function(pDelegate, aIID) {
    	            return pDelegate.QueryInterface(CI.nsIDOMWindowInternal).
    	                QueryInterface(CI.nsIInterfaceRequestor).getInterface(CI.nsIWebNavigation);
    	        } );
    	case "Supports_QueryInterface":
    	    return new RFComponentManager(
    	        function(aContractID, aDelegate, aIID) {
    	            return pDelegate.QueryInterface(CI[aContractID]);
    	        } );
    	case "DOMXULElement_GetBoxObject": return RFSimpleFactory(CI.nsIDOMXULElement, "boxObject");
    	case "HTMLElement_GetOffsetParent": return RFSimpleFactory(CI.nsIDOMNSHTMLElement, "offsetParent");
    	case "WebNavigation_GetDocument": return RFSimpleFactory(CI.nsIWebNavigation, "document");
    	case "WebNavigation_GetCurrentURI": return RFSimpleFactory(CI.nsIWebNavigation, "currentURI");
    	case "WindowMediator_GetEnumerator": //not used first arg
    	    return new RFFactory(
    	        function(pDelegate, aIID) {
    	            return Components.classes["@mozilla.org/appshell/window-mediator;1"].
        	            getService(CI.nsIWindowMediator).getEnumerator("navigator:browser");
    	        } );
    	case "DocShellTreeItem_GetTreeOwner": return RFSimpleFactory(CI.nsIDocShellTreeItem, "treeOwner");
    	case "DocShellTreeItem_GetRootTreeItem": return RFSimpleFactory(CI.nsIDocShellTreeItem, "rootTreeItem");
    	case "DocShellTreeItem_GetSameTypeParent": return RFSimpleFactory(CI.nsIDocShellTreeItem, "sameTypeParent");
    	case "Document_GetFocusedElement":
    	    return new RFFactory(
    	        function(pDelegate, aIID) {
    	            //need better solution here
    	            return pDelegate.defaultView.QueryInterface(Components.interfaces.nsIDOMWindowInternal).
                        QueryInterface(Components.interfaces.nsIInterfaceRequestor).
                        getInterface(Components.interfaces.nsIWebNavigation).
                        QueryInterface(Components.interfaces.nsIDocShell).chromeEventHandler.ownerDocument.
                        commandDispatcher.focusedElement;
    	        } );
    	case "XULWindow_GetInterfaceGetDocShell":
    	    return new RFFactory(
    	        function(pDelegate, aIID) {
    	            return pDelegate.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
    	                .getInterface(Components.interfaces.nsIXULWindow)
    	                .docShell;
    	        } );
    	
        //RFSimpleLoadGroup
        case "Window_GetInnerHeight": return RFSimpleArray(CI.nsIDOMWindowInternal, "innerHeight");
        case "Window_GetInnerWidth": return RFSimpleArray(CI.nsIDOMWindowInternal, "innerWidth");
        case "Window_GetLocation":
            return RFSimpleLoadGroup(CI.nsIDOMWindowInternal, CI.nsISupportsString, "location");
        case "Window_GetFullScreen": return RFSimpleArray(CI.nsIDOMWindowInternal, "fullScreen");
        case "DOMXULElement_GetHidden": return RFSimpleArray(CI.nsIDOMXULElement, "hidden");
        case "DOMXULElement_GetCollapsed": return RFSimpleArray(CI.nsIDOMXULElement, "collapsed");
        case "BoxObject_GetScreenX": return RFSimpleArray(CI.nsIBoxObject, "screenX");
        case "BoxObject_GetScreenY": return RFSimpleArray(CI.nsIBoxObject, "screenY");
        case "BoxObject_GetHeight": return RFSimpleArray(CI.nsIBoxObject, "height");
        case "BoxObject_GetWidth": return RFSimpleArray(CI.nsIBoxObject, "width");
        case "HTMLElement_GetInnerHTML":
            return RFSimpleLoadGroup(CI.nsIDOMNSHTMLElement, CI.nsISupportsString, "innerHTML");
        case "HTMLElement_GetClientHeight": return RFSimpleArray(CI.nsIDOMNSElement, "clientHeight");
        case "HTMLElement_GetClientWidth": return RFSimpleArray(CI.nsIDOMNSElement, "clientWidth");
        case "HTMLElement_GetClientTop": return RFSimpleArray(CI.nsIDOMNSElement, "clientTop");
        case "HTMLElement_GetClientLeft": return RFSimpleArray(CI.nsIDOMNSElement, "clientLeft");
        case "HTMLElement_GetOffsetLeft": return RFSimpleArray(CI.nsIDOMNSHTMLElement, "offsetLeft");
        case "HTMLElement_GetOffsetHeight": return RFSimpleArray(CI.nsIDOMNSHTMLElement, "offsetHeight");
        case "HTMLElement_GetOffsetWidth": return RFSimpleArray(CI.nsIDOMNSHTMLElement, "offsetWidth");
        case "HTMLElement_GetOffsetTop": return RFSimpleArray(CI.nsIDOMNSHTMLElement, "offsetTop");
        case "DOMKeyEvent_GetKeyCode": return RFSimpleArray(CI.nsIDOMKeyEvent, "keyCode");
        case "DOMKeyEvent_GetCharCode": return RFSimpleArray(CI.nsIDOMKeyEvent, "charCode");
        case "DOMKeyEvent_GetAltKey": return RFSimpleArray(CI.nsIDOMKeyEvent, "altKey");
        case "DOMKeyEvent_GetCtrlKey": return RFSimpleArray(CI.nsIDOMKeyEvent, "ctrlKey");
        case "DOMKeyEvent_GetShiftKey": return RFSimpleArray(CI.nsIDOMKeyEvent, "shiftKey");
        case "DOMKeyEvent_QI": 
            return new RFArray( function(ulIndex, aElem) {
                try {
                    return !!aElem.QueryInterface(CI.nsIDOMKeyEvent);
                } catch (e) {
                    return  0;
                }
    	    } );
        case "Utils_ForceGC": 
            return new RFArray( function(ulIndex, aElem) {
                Components.utils.forceGC();
    	    } );
        case "DocShell_GetBusyFlags": return RFSimpleArray(CI.nsIDocShell, "busyFlags");
        case "DocShellTreeNode_GetChildCount": return RFSimpleArray(CI.nsIDocShellTreeNode, "childCount");
        case "WebNavigation_Reload":
            return new RFArray( function(ulIndex, aElem) {
                aElem.QueryInterface(CI.nsIWebNavigation).reload(ulIndex);
    	    } );
        case "DocShell_QI":
            return new RFArray( function(ulIndex, aElem) {
                try {
                    return !!aElem.QueryInterface(CI.nsIDocShell);
                } catch (e) {
                    return  0;
                }
    	    } );
        case "RDF_IsToolbarCollapsed":
            return new RFArray( function(ulIndex, aElem) {
                var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(CI.nsIRDFService);
    			var localStore = rdfService.GetDataSource("rdf:local-store");
			    var collapsedRes = rdfService.GetResource("collapsed");
			    var rfbarRes = rdfService.GetResource("chrome://browser/content/browser.xul#roboform-bar");
    			return localStore.GetTarget(rfbarRes, collapsedRes, true).
    			    QueryInterface(Components.interfaces.nsIRDFLiteral).Value == "true";
    	    } );

        //RFLoadGroup
        case "WebNavigation_LoadURI":
            return new RFLoadGroup( function(aRequest, aElem) {
                var callback = aRequest.QueryInterface(CI.nsISupportsString);
                aElem.QueryInterface(CI.nsIWebNavigation).
                    loadURI(callback.data, callback.type, null,null,null);
    	    } );
        case "DOMDocumentXBL_GetAnonymousElementByAttributeInput":
            return new RFLoadGroup( function(aRequest, aElem) {
                var callback = aRequest.QueryInterface(CI.nsISupportsInterfacePointer);
                callback.data = aElem.QueryInterface(CI.nsIDOMDocumentXBL).getAnonymousElementByAttribute(
                    callback.data, "anonid", "input");
    	    } );
        case "DOMDocumentXBL_GetAnonymousElementByAttributeAccept":
            return new RFLoadGroup( function(aRequest, aElem) {
                var callback = aRequest.QueryInterface(CI.nsISupportsInterfacePointer);
                callback.data = aElem.QueryInterface(CI.nsIDOMDocumentXBL).getAnonymousElementByAttribute(
                    callback.data, "dlgtype", "accept");
    	    } );
        case "DocShellTreeItem_IsContentType":
            return new RFArray( function(ulIndex, aElem) {
                return aElem.QueryInterface(CI.nsIDocShellTreeItem).itemType == CI.nsIDocShellTreeItem.typeContent;
    	    } );
        case "DocShell_IsTab":
            return new RFArray( function(ulIndex, aElem) {
                try {
                    var browsers = aElem.QueryInterface(CI.nsIDocShellTreeItem).rootTreeItem
                        .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                        .getInterface(Components.interfaces.nsIDOMWindow)
                        .getBrowser().browsers;
                    for(var i = 0; i < browsers.length; i++)
                        if (browsers[i].docShell == aElem)
                            return true;
                    return false;
                } catch (e) {
                    return false;
                }
    	    } );
        case "DOMViewCSS_GetComputedStyle":
            return new RFLoadGroup( function(aRequest, aElem) {
                var callback = aRequest.QueryInterface(CI.nsISupportsInterfacePointer);
                callback.data = aElem.QueryInterface(CI.nsIDOMViewCSS).getComputedStyle(callback.data, '');
    	    } );
        case "DocShellTreeNode_GetChildAt":
            return new RFLoadGroup( function(aRequest, aElem) {
                var callback = aRequest.QueryInterface(CI.nsISupportsInterfacePointer);
                callback.data = aElem.QueryInterface(CI.nsIDocShellTreeNode).getChildAt(callback.type);
    	    } );
        case "WebNavigation_GoBack": //need only nsISupports
            return new RFLoadGroup( function(aRequest, aElem) {
                aElem.QueryInterface(CI.nsIWebNavigation).goBack(); 
            } );
        case "WebNavigation_Stop": //need only nsISupports
            return new RFLoadGroup( function(aRequest, aElem) {
                aElem.QueryInterface(CI.nsIWebNavigation).stop(CI.nsIWebNavigation.STOP_ALL);
            } );
        case "WebNavigation_GoForward": //need only nsISupports
            return new RFLoadGroup( function(aRequest, aElem) {
                aElem.QueryInterface(CI.nsIWebNavigation).goForward(); 
            } );
        case "Document_FlushPendingNotifications": //need only nsISupports
            return new RFLoadGroup( function(aRequest, aElem) {
                aElem.QueryInterface(CI.nsIDOMNSDocument).elementFromPoint(0,0); 
            } );
        case "DOMXULElement_Click": //not using callback
            return new RFLoadGroup( function(aRequest, aElem) {
                aElem.QueryInterface(CI.nsIDOMXULElement).click(); 
            } );
        case "KeyEvent_InitKeyPressReturn": //not using callback
            return new RFLoadGroup( function(aRequest, aElem) {
                aElem.QueryInterface(CI.nsIDOMKeyEvent).initKeyEvent("keypress", true, true, null,
                    /*ctrl*/false, /*alt*/false, /*shift*/false, 0, CI.nsIDOMKeyEvent.DOM_VK_RETURN, 0);
            } );
    	    
    	//RFUnicharStreamListener
        case "Window_Navigate":
            return new RFUnicharStreamListener( function(aRequest, aContext, aData) {
                aContext.QueryInterface(nsIDOMWindowInternal).location.assign(aData);
    	    } );
    	case "KeyEvent_InitKeyPress":
            return RFStreamListener( function(aRequest, aContext, aInputStream, aOffset, aCount) {
                aContext.QueryInterface(CI.nsIDOMKeyEvent).initKeyEvent("keypress", true, true, null,
                    /*ctrl*/aOffset&1, /*alt*/aOffset&2, /*shift*/aOffset&4, 0, 0, aCount);
    	    } );
    	case "KeyEvent_InitKeyDown":
            return RFStreamListener( function(aRequest, aContext, aInputStream, aOffset, aCount) {
                aContext.QueryInterface(CI.nsIDOMKeyEvent).initKeyEvent("keypress", true, true, null,
                    /*ctrl*/aOffset&1, /*alt*/aOffset&2, /*shift*/aOffset&4, 0, aCount, 0);
    	    } );
    	case "KeyEvent_InitKeyUp":
            return RFStreamListener( function(aRequest, aContext, aInputStream, aOffset, aCount) {
                aContext.QueryInterface(CI.nsIDOMKeyEvent).initKeyEvent("keypress", true, true, null,
                    /*ctrl*/aOffset&1, /*alt*/aOffset&2, /*shift*/aOffset&4, 0, aCount, 0);
    	    } );
    	
function RFStreamListener(handler) {
    return new RFCommonWraper("onDataAvailable", handler, CI.nsIStreamListener);
}
//    void onDataAvailable(in nsIRequest aRequest,
//                         in nsISupports aContext, 
//                         in nsIInputStream aInputStream,
//                         in unsigned long aOffset,
//                         in unsigned long aCount);

    	default:
    	    NOT_IMPLEMENTED();
    	}
    },
    
	// nsISupports
	QueryInterface: function QueryInterface(aIID) {
		if (aIID.equals(nsIFactory) ||
			aIID.equals(nsIComponentManager) ||
			aIID.equals(nsISupports))
			return this;
		throw Components.results.NS_ERROR_NO_INTERFACE;
		return null;
	}
};

const Module = {
  // nsIModule
  unregisterSelf: function(compMgr, location, type)
  {
    compReg = compMgr.QueryInterface(nsIComponentRegistrar);
    compReg.unregisterFactoryLocation(NS_RFHELPER_CID, location);        
  },

  registerSelf: function registerSelf(compMgr, fileSpec, location, type) {
    compReg = compMgr.QueryInterface(nsIComponentRegistrar);

    compReg.registerFactoryLocation(NS_RFHELPER_CID,
                                    "nsRFHelper",
                                    NS_RFHELPER_CONTRACTID,
                                    fileSpec,
                                    location,
                                    type);
  },

  // nsIModule
  getClassObject: function getClassObject(compMgr, aCID, aIID) {
    if (!aIID.equals(nsIFactory)) {
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    }
    if (aCID.equals(NS_RFHELPER_CID)) {
      return RFHelper;
    }

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  // nsIModule
  canUnload: function canUnload() {
    return true;
  },

  // nsISupports
  QueryInterface: function QueryInterface(aIID) {
    if (aIID.equals(nsIModule) ||
       (aIID.equals(nsISupports)))
      return this;

    throw Components.results.NS_ERROR_NO_INTERFACE;
    return null;
  }
};

function NSGetModule(compMgr, fileSpec) {
  return Module;
}
